/** LANL:license
 * -------------------------------------------------------------------------
 * This SOFTWARE has been authored by an employee or employees of the
 * University of California, operator of the Los Alamos National Laboratory
 * under Contract No. W-7405-ENG-36 with the U.S. Department of Energy.
 * The U.S. Government has rights to use, reproduce, and distribute this
 * SOFTWARE.  The public may copy, distribute, prepare derivative works and
 * publicly display this SOFTWARE without charge, provided that this Notice
 * and any statement of authorship are reproduced on all copies.  Neither
 * the Government nor the University makes any warranty, express or implied,
 * or assumes any liability or responsibility for the use of this SOFTWARE.
 * If SOFTWARE is modified to produce derivative works, such modified
 * SOFTWARE should be clearly marked, so as not to confuse it with the
 * version available from LANL.
 * -------------------------------------------------------------------------
 * LANL:license
 * -------------------------------------------------------------------------
 */
/**
 * xmlgen : XML generator for component creation tools based on PDToolkit
 *
 * Written by Matthew Sottile (matt@lanl.gov)
 *
 * Contact: Matthew Sottile, MS B287
 *          Advanced Computing Laboratory,
 *          Los Alamos National Laboratory,
 *          Los Alamos, NM 87545
 *
 * Conforms to DTD specified in mapping.dtd included with this code.
 */
#include <unistd.h>
#include "pdb.h"
#include "pdbAll.h"
#include <vector>
#include <iostream>
#include <fstream>

using namespace std;

//
// enumeration for which language we're dealing with.  I'd rather not
// depend on pdbTypes:: enumerations for this kind of thing, so we have our
// own here with JUST the languages currently supported.
//
enum language_t { LANG_FORTRAN=1, 
		  LANG_C=2, 
		  LANG_CXX=3 };

// 
// global flag with the language
//
language_t libraryLanguage;

//
// global vector of fully-qualified names of subroutines that have
// been output already.  since ductape doesn't have a way to ask for
// globals, we have to make sure that modules/namespaces/classes have
// been dumped FIRST, and then go and figure out what stuff hasn't
// come out yet and see if they qualify as valid globals.
//
vector<string> dumpedRoutines;
vector<string> dumpedTypes;
vector<string> dumpedClasses;

//
// macro for indentation
//
#define indent(os,depth) { for (int i=0;i<depth;i++) os << "  "; }

//
// predefine these guys since we need them and there's a bit of a 
// circular dependency with outputTypeKinds.
//
int outputArray(const pdbType *ty, ostream& os, int depth);
int outputIndirect(const pdbType *ty, ostream& os, int depth);

//
// code to print type kind info -- based on code in pdbType.cc::printType
//
int outputTypeKinds(const pdbType *_ty, ostream& os, int depth) {
  const pdbType *ty = _ty;

  // 
  // we shouldn't see this.
  // if we do, dump a comment to mark the spot in the output and
  // send the error to STDERR
  //
  if (ty == NULL) {
    os << "<!-- OUTPUTTYPEKINDS ERROR OCCURRED HERE --> " << endl;
    cerr << "ERROR: attempting to print null type." << endl;
    return 1;
  }

  //
  // kind is dumped as is -- specialized types are conditionally
  // printed below
  //
  // NOTE: examine group to see if kind is actually a user defined
  //       type
  //

  //
  // deal with TREF stuff
  //
  if (ty->kind() == pdbItem::TY_TREF) {
    ty = ty->baseType();
  }

  const pdbGroup *cd = ty->isGroup();

  indent(os,depth);
  os << "<type ";

  // 
  // if ty is a group, then we treat it as a usertype and dump the
  // appropriate name.  otherwise, dump the ductape provided name.
  //
  if (cd == NULL)
    os << "kind=\"" << ty->toName(ty->kind()) << "\" ";
  else {
    //
    // if the parent group is not null, we can find out where the
    // type is defined.  For example, if this is a fortran user
    // defined type, then the parent group is the module in which
    // the type is declared.
    //
    if (cd->parentGroup() != NULL)
      os << "kind=\"usertype\" usertype=\"" << cd->name() << "\" "
	 << "context=\"" << cd->parentGroup()->name() << "\" ";
    else
      os << "kind=\"usertype\" usertype=\"" << cd->name() << "\" ";
  }

  //
  // see printType method mentioned above for a very similar switch
  //
  switch(ty->kind()) {
  case pdbItem::TY_INT:
  case pdbItem::TY_BOOL:
  case pdbItem::TY_WCHAR:
  case pdbItem::TY_FINT:
  case pdbItem::TY_FLOGIC:
  case pdbItem::TY_FCHAR:
  case pdbItem::TY_ENUM:
    os << "ikind=\"" << ty->toName(ty->integerType())
       << "\" ";
    break;
  case pdbItem::TY_FLOAT:
  case pdbItem::TY_FFLOAT:
  case pdbItem::TY_FCMPLX:
    os << "fkind=\"" << ty->toName(ty->floatType()) << "\" ";
    break;
  case pdbItem::TY_TPARAM:
    //
    // template junk
    //
    os << "genericname=\"" << ty->name() << "\" ";
    break;
  default:
    break;
  }

  //
  // we have a string, so dump the length
  // I think this only applies to FCHAR
  //
  if (ty->kind() == pdbItem::TY_FCHAR) {
    if (ty->numElements() == -1)
      os << "clen=\"*\" ";
    else
      os << "clen=\"" << ty->numElements() << "\" ";
  }

  os << ">" << endl;

  //
  // if the type is an array or indirection, we need to emit those elements
  // in the context of the currently open TYPE
  //
  outputArray(ty,os,depth+1);
  outputIndirect(ty,os,depth+1);
  
  //
  // close tag
  //
  indent(os,depth);
  os << "</type>" << endl;
  return 0;
}

//
// code to deal with pointers and references
//
int outputIndirect(const pdbType *ty, ostream& os, int depth) {
  int ptrLevels = 0;

  if ((ty->kind() == pdbItem::TY_PTR) ||
      (ty->kind() == pdbItem::TY_FPTR)) {
    //
    // deal with pointers
    //
    ptrLevels++;

    const pdbType *ptype = ty->pointedToType();
    
    //
    // iterate down the pointedToType chain to figure out how deep the
    // pointers go.
    //
    while (ptype != NULL && (ptype->kind() == pdbItem::TY_PTR ||
                             ptype->kind() == pdbItem::TY_FPTR)) {
      ptrLevels++;
      ptype = ptype->pointedToType();
    }

    indent(os,depth);
    os << "<indirect depth=\"" << ptrLevels << "\" kind=\"pointer\">" << endl; 

    //
    // dump the type that this pointer points at
    //
    outputTypeKinds(ptype,os,depth+1);

    indent(os,depth);
    os << "</indirect>" << endl;
  } else if (ty->kind() == pdbItem::TY_REF) {
    //
    // deal with references
    //
    ptrLevels++;

    const pdbType *rtype = ty->referencedType();
    
    indent(os,depth);
    os << "<indirect depth=\"" << ptrLevels << "\" kind=\"reference\">" 
       << endl;

    outputTypeKinds(rtype,os,depth+1);

    indent(os,depth);
    os << "</indirect>" << endl;
  }

  return 0;
}

//
// code to print array elements
//
int outputArray(const pdbType *ty, ostream& os, int depth) {
  bool explicitShape = false;
  bool isFortran = false;

  if ((ty->kind() == pdbItem::TY_ARRAY) || 
      (ty->kind() == pdbItem::TY_FARRAY)) {
    //
    // check shape
    //
    if (ty->shape() == pdbItem::SH_EXPLICIT) {
      explicitShape = true;
    } 

    //
    // print array element: element kind, rank, language
    // the rank only seems to work for FARRAY
    //
    indent(os,depth);
    os << "<array ";
    if (ty->kind() == pdbItem::TY_FARRAY) {
      os << "lang=\"fortran\" ";
      os << "rank=\"" << ty->rank() << "\">" << endl;
      isFortran = true;
    } else {
      //
      // CRank is 1.  Looking at rank() is bogus
      //
      os << "lang=\"c\" rank=\"1\">" << endl;
    }
    
    outputTypeKinds(ty->elementType(),os,depth+1);

    // 
    // c-arrays are really braindead - they don't even show
    // dimension info.  Grrr.
    //
    if (ty->kind() == pdbItem::TY_ARRAY) {
      indent(os,depth+1);
      os << "<dimension extent=\"*\"/>" << endl;
    }

    //
    // iterate over dimensions
    // eventually we will probably have better information from
    // the pdbBounds class, and can then stop the INT_MAX stupidity.
    //
    const pdbType::boundvec bv = ty->bounds();
    for (pdbType::boundvec::const_iterator b = bv.begin();
	 b != bv.end();
	 ++b) {
      const pdbBounds curb = static_cast<const pdbBounds>(*b);
      indent(os,depth+1);

      //
      // if upper == INT_MAX, upper is undefined.  Very dumb hack IMHO.
      // if upper bound was NA, then the array should not be tagged
      // as having explicit shape (right?) -mjs 7/24/02
      //
      if ((&curb)->upper() == INT_MAX)
        explicitShape = false;
        
      //
      // if explicitShape, we know extent and lower_bound (if non-zero or one)
      // otherwise, no lower bound specified and the extent is set to
      // "*".
      //
      if (explicitShape) {
        os << "<dimension extent=\"" << (&curb)->upper()-(&curb)->lower()+1
           << "\" ";
        if (( (&curb)->lower() != 0 && !isFortran) ||
            ( (&curb)->lower() != 1 &&  isFortran) )
          os << "lower_bound=\"" << (&curb)->lower() << "\"";
        os << "/>" << endl;
      } else {
        os << "<dimension extent=\"*\"/>" << endl;
      }
    }

    //
    // emit close tag
    //
    indent(os,depth);
    os << "</array>" << endl;
  }

  return 0;
}

//
// dump a single pdbArg out to os
//
int outputArgument(const pdbArg *arg, ostream& os, int depth) {
  //
  // comment block
  //
  os << endl;
  indent(os,depth);
  os << "<!-- ARGUMENT : " << arg->name() << " -->" << endl;

  //
  // begin argument
  //
  indent(os,depth);
  os << "<arg name=\"" << arg->name() << "\" ";
  
  //
  // dump intent
  //  
  if (arg->intentOut()) {
    if (arg->intentIn()) {
      os << "intent=\"inout\">";
    } else {
      os << "intent=\"out\">";
    }
  } else {
    os << "intent=\"in\">";
  }

  //
  // if it is an array, we dump that here.
  //
  os << endl;

  //
  // print specific type kind information
  //
  outputTypeKinds(arg->type(),os,depth+1);

  //
  // end arg tag
  //
  indent(os,depth);
  os << "</arg>" << endl;

  return 0;
}

//
// Output a single fortran routine
//
int outputFRoutine(const pdbFRoutine *cur, ostream& os, int depth) {
  const pdbType *sig = cur->signature();

  //
  // note: not sure what sig==NULL implies about the function itself.
  //
  if (sig != NULL) {
    const pdbLoc ploc = cur->firstExecStmtLocation();

    // 
    // if we have no known implementation, this is 0 and we shouldn't generate
    // a signature.  Wrapper generators aren't guaranteed an implementation
    // exists, so we don't want to lead them into believing they should
    // worry about wrapper code
    //
    if (ploc.line() == 0) {
      dumpedRoutines.push_back(cur->fullName());
      return 0;
    }

    //
    // dump the routine
    //
    indent(os,depth);
    
    // 
    // comment block
    //
    os << endl;
    indent(os,depth);
    os << "<!-- " << endl;
    indent(os,depth);
    os << "  -  METHOD: " << cur->fullName() << endl;
    indent(os,depth);
    os << "  -->" << endl;
    indent(os,depth);
    dumpedRoutines.push_back(cur->fullName());
    
    //
    // begin method
    //
    os << "<method name=\"" << cur->name() << "\" ";

    const pdbFRoutine *isAlias = cur->isAliasFor();
    if (isAlias != NULL) {
      os << "alias=\"" << isAlias->fullName() << "\" ";
    }

    //
    // aliases (methods from INTERFACE blocks) are considered abstract.
    // otherwise, fortran subroutines and functions are considered static.
    //
    if (cur->kind() == pdbType::RO_FALIAS) {
      os << "kind=\"abstract\">" << endl;
    } else {
      os << "kind=\"static\">" << endl;
    }

    //
    // dump each argument
    //
    const pdbType::argvec av = sig->arguments();
    for (pdbType::argvec::const_iterator a = av.begin();
	 a != av.end();
	 ++a) {
      const pdbArg cur = static_cast<const pdbArg>(*a);
      outputArgument(&cur,os,depth+1);
    }

    //
    // dump the return value
    //
    const pdbType *ret = sig->returnType();
    os << endl;
    indent(os,depth+1);
    os << "<!-- return value -->" << endl;
    indent(os,depth+1);
    os << "<return>" << endl;

    outputTypeKinds(ret,os,depth+2);
    
    indent(os,depth+1);
    os << "</return>" << endl;
    
    //
    // close method
    //
    indent(os,depth);
    os << "</method>" << endl;
  } else {
    //
    // SIG == NULL means we might be dealing with a function alias here
    // 
    dumpedRoutines.push_back(cur->fullName());

    if (cur->kind() == pdbItem::RO_FALIAS) {
      // 
      // comment block
      //
      os << endl;
      indent(os,depth);
      os << "<!-- " << endl;
      indent(os,depth);
      os << "  -  PROCEDUREMAP: " << cur->fullName() << endl;
      indent(os,depth);
      os << "  -->" << endl;
      indent(os,depth);
      
      os << "<proceduremap name=\"" << cur->fullName() << "\">" << endl;
      //
      // dump the implementations of the generic as concrete elements
      //
      const pdbFRoutine::implvec iv = cur->implementedBy();
      for (pdbFRoutine::implvec::const_iterator i = iv.begin();
           i != iv.end();
           ++i) {
        pdbFRoutine *fr = static_cast<pdbFRoutine *>(*i);
        indent(os,depth+1);
        os << "<concrete name=\"" << fr->name() << "\"/>" << endl;
      }

      indent(os,depth);
      os << "</proceduremap>" << endl << endl;

    } else {
      cerr << "METHOD: " << cur->fullName() << " has null signature." << endl;
      cerr << "        method not emitted to XML." << endl;
    }
  }

  return 0;
}

/**
 * This method outputs a user-defined fortran type.
 */
int outputFUserType(const pdbGroup *gr, ostream& os, int depth) {
  os << endl;
  indent(os,depth);
  os << "<!--" << endl;
  indent(os,depth);
  os << "  - User defined type : " << gr->name() << endl;
  indent(os,depth);
  os << "  -->" << endl;

  dumpedTypes.push_back(gr->fullName());

  indent(os,depth);
  os << "<structure name=\"" << gr->name() << "\" kind=\"ftype\">" << endl;

  //
  // iterate over fields
  //
  pdbGroup::fieldvec typefields = gr->dataMembers();
  for (pdbGroup::fieldvec::const_iterator tf = typefields.begin();
       tf != typefields.end();
       ++tf) {
    const pdbGroupField *tfield =
      static_cast<const pdbGroupField*>(*tf);

    os << endl;
    indent(os,depth+1);
    os << "<!-- FIELD: " << tfield->name() << " -->" << endl;
    indent(os,depth+1);
    os << "<field name=\"" << tfield->name() << "\">" << endl;

    outputTypeKinds(tfield->type(),os,depth+2);

    indent(os,depth+1);
    os << "</field>" << endl;
  }
  
  indent(os,depth);
  os << "</structure>" << endl;

  return 0;
}

//
// Output a single C routine
//
int outputCRoutine(const pdbCRoutine *cur, ostream& os, int depth) {
  const pdbType *sig = cur->signature();

  //
  // bail out on compiler generated stuff
  //
  if (cur->isCompilerGenerated()) return 0;

  //
  // note: not sure what sig==NULL implies about the function itself.
  //
  if (sig != NULL) {

    //
    // dump the routine
    //
    indent(os,depth);
    
    // 
    // comment block
    //
    os << endl;
    indent(os,depth);
    os << "<!-- " << endl;
    indent(os,depth);
    os << "  -  METHOD: " << cur->fullName() << endl;
    indent(os,depth);
    os << "  -->" << endl;
    indent(os,depth);
    dumpedRoutines.push_back(cur->fullName());
    
    //
    // begin method
    //
    os << "<method name=\"" << cur->name() << "\" ";
    
    if (cur->isVirtual())
      os << "kind=\"virtual\">" << endl;
    else if (cur->isStatic())
      os << "kind=\"static\">" << endl;
    else
      os << "kind=\"final\">" << endl;

    //
    // dump each argument
    //
    const pdbType::argvec av = sig->arguments();
    for (pdbType::argvec::const_iterator a = av.begin();
	 a != av.end();
	 ++a) {
      const pdbArg cur = static_cast<const pdbArg>(*a);
      outputArgument(&cur,os,depth+1);
    }

    //
    // dump the return value
    //
    const pdbType *ret = sig->returnType();
    os << endl;
    indent(os,depth+1);
    os << "<!-- return type -->" << endl;
    indent(os,depth+1);
    os << "<return>" << endl;
    outputTypeKinds(ret,os,depth+2);
    indent(os,depth+1);
    os << "</return>" << endl;
    
    //
    // close method
    //
    indent(os,depth);
    os << "</method>" << endl;
  } else {
    //
    // back to this sig==NULL weirdness.  Currently, do NOT dump, just warn.
    //
    dumpedRoutines.push_back(cur->fullName());

    cerr << "METHOD: " << cur->fullName() << " has null signature." << endl;
    cerr << "        method not emitted to XML." << endl;
  }

  return 0;
}

/**
 * this method outputs C/C++ routines not contained in classes or other
 * containers.
 */
int outputCRoutines(PDB::croutinevec& crv, ostream& os, int depth) {
  bool dumped = false;
  bool printedScope = false;

  //
  // iterate over each routine, outputting the XML for each
  //
  for (PDB::croutinevec::const_iterator c = crv.begin();
       c != crv.end();
       ++c) {
    const pdbCRoutine *cur = static_cast<const pdbCRoutine*>(*c);

    //
    // check if it has already been dumped...
    //
    for (unsigned int i = 0; i < dumpedRoutines.size(); i++) {
      if ((string)dumpedRoutines[i] == cur->fullName()) {
        dumped = true;
        break;
      }
    }

    //
    // if a method has a null signature, we definitely don't dump it.
    // so pretend it has already been dumped.
    //
    const pdbType *sig = cur->signature();
    if (sig == NULL)
      dumped = true;

    //
    // compiler generated functions don't get dumped either.  (things
    // like operator= and other junk.)  I hate C++!
    //
    if (cur->isCompilerGenerated() == true)
      dumped = true;

    if (dumped == false) {
      if (printedScope == false) {
        indent(os,depth);
        os << "<scope name=\"CXXGlobal\" kind=\"global\">" <<
          endl;
        printedScope = true;
      }
      outputCRoutine(cur,os,depth+1);
      dumpedRoutines.push_back(cur->fullName());
    }

    dumped = false;
  }

  if (printedScope == true) {
    indent(os,depth);
    os << "</scope>" << endl;
  }
  
  return 0;
}

/**
 * dump template info for debug info.
 */
void dumpTemplate(const pdbTemplate *ctempl) {
  if (ctempl == NULL) return;

  cerr << "DUMPING TEMPLATE:" << endl;
  cerr << "desc() = " << ctempl->desc() << endl;
  cerr << "text() = " << ctempl->text() << endl;
  cerr << "name() = " << ctempl->name() << endl;

  cerr << "==>declaration" << endl;
  dumpTemplate(ctempl->declaration());

  cerr << "==>definition" << endl;
  dumpTemplate(ctempl->definition());

  cerr << "---->template arguments" << endl;
  const pdbTemplate::targvec &tv = ctempl->arguments();
  for (pdbTemplate::targvec::const_iterator ta = tv.begin();
       ta != tv.end();
       ++ta) {
    const pdbTemplateArg targ =
      static_cast<const pdbTemplateArg>(*ta);

    cerr << "-->name() = " << targ.name() << endl;
    cerr << "-->kind() = " << targ.toName(targ.kind()) << endl;
    cerr << "-->value() = " << targ.value() << endl;
    cerr << "-->defaultValue() = " << targ.defaultValue() << endl;
    cerr << "-->isSpecialization() = " << targ.isSpecialization() << endl;

    cerr << "---->TEMPLATEARG()" << endl;
    dumpTemplate(targ.templateArg());
    cerr << "---->DEFAULTTEMPLATEARG()" << endl;
    dumpTemplate(targ.defaultTemplateArg());
    
    const pdbType *t = targ.type();
    cerr << "---->TYPE() name=" << t->name() << endl;
    cerr << "---->       kind=" << t->toName(t->kind()) << endl;
    
  }

}

/**
 * This method outputs a C++ class or C struct. 
 */
int outputClass(const pdbClass *c, ostream &os, int depth) {

  os << endl;
  indent(os,depth);
  os << "<!--" << endl;
  indent(os,depth);
  if (libraryLanguage == LANG_C)
    os << "   - STRUCTURE: " << c->fullName() << endl;
  else
    os << "   - CLASS: " << c->fullName() << endl;
  indent(os,depth);
  os << "  -->" << endl;

  dumpedClasses.push_back(c->fullName());

  //
  // print open tag for class structure
  //
  indent(os,depth);
  if (libraryLanguage == LANG_CXX) {
    os << "<structure kind=\"cxxclass\" name=\"" << c->name() << "\" ";
    //
    // get base class vector.  For now, we will just put the first parent
    // as the parent.  Multiple inheritance is stupid anyways.  (And don't
    // waste your breath arguing with me on that point - lots of languages
    // prove that it's not necessary).
    //
    const pdbClass::basevec &bv = c->baseClasses();
    if (bv.size() != 0) {
      pdbBase *base = static_cast<pdbBase *>(bv[0]);
      const pdbClass *baseclass = base->base();
      os << "parent=\"" << baseclass->fullName() << "\" ";
    }
  } else
    os << "<structure kind=\"cstruct\" name=\"" << c->name() << "\" ";

  const pdbTemplate *ctempl = c->isTemplate();

  //
  // ... so its templated.
  //
  if (ctempl != NULL) {
    // 
    // switch on the different template types
    //
    switch (ctempl->kind()) {
    case pdbItem::TE_CLASS:
      os << " templated=\"yes\" ";
      dumpTemplate(ctempl);

      //os << "<!-- " << ctempl->text() << " --> " << endl;
      //      const pdbTemplateItem::targvec *targs = c->speclArguments();
      //for (pdbTemplateItem::targvec::const_iterator ta = targs->begin();
      //     ta != targs->end();
      //     ++ta) {
      //  const pdbTemplateArg ti =
      //    static_cast<const pdbTemplateArg>(*ta);
      //  
      //  os << "<!-- template arg: " << ti.name() << " -->" << endl;
      //}
      
      break;
    default:
      cerr << "Unsupported template type found in class '" <<
        c->name() << "'" << endl;
    }
  }

  //
  // close element open tag
  //
  os << ">" << endl;

  // now go through fields and methods
  // fields first.
  const pdbGroup::fieldvec fv = ((const pdbGroup *)c)->dataMembers();
  for (pdbGroup::fieldvec::const_iterator tf = fv.begin();
       tf != fv.end();
       ++tf) {
    const pdbGroupField *tfield =
      static_cast<const pdbGroupField*>(*tf);

    os << endl;
    indent(os,depth+1);
    os << "<!-- FIELD: " << tfield->name() << " -->" << endl;
    indent(os,depth+1);
    os << "<field name=\"" << tfield->name() << "\">" << endl;

    outputTypeKinds(tfield->type(),os,depth+2);

    indent(os,depth+1);
    os << "</field>" << endl;
  }

  // then methods
  const pdbClass::methodvec mv = c->methods();
  for (pdbClass::methodvec::const_iterator m = mv.begin();
       m != mv.end();
       ++m) {
    const pdbMethod *meth =
      static_cast<const pdbMethod *>(*m);
    outputCRoutine(meth->func(),os,depth+1);
  }  

  //
  // close structure
  //
  indent(os,depth);
  os << "</structure>" << endl;

  return 0;
}

/**
 * This method outputs all C++ namespaces
 */
int outputNamespaces(PDB::namespacevec& nv, ostream& os, int depth) {
  for (PDB::namespacevec::const_iterator n = nv.begin();
       n != nv.end();
       ++n) {
    const pdbNamespace *cur = static_cast<const pdbNamespace*>(*n);
    pdbNamespace::memvec mv = cur->members();

    // 
    // make sure the namespace isn't empty - those are useless and stupid.
    //
    if (mv.begin() != mv.end()) {
      //
      // comment the XML
      //
      os << endl;
      indent(os,depth);
      os << "<!--" << endl;
      indent(os,depth);
      os << "   - Namespace: " << cur->fullName() << endl;
      indent(os,depth);
      os << "  -->" << endl << endl;

      //
      // begin scope tag
      //
      indent(os,depth);
      os << "<scope name=\"" << cur->fullName() << "\" kind=\"namespace\"" <<
	">" << endl;

      //
      // iterate over namespace members
      // (how one does this is not obvious...  more reason PDT is
      //  lightyears from user friendly...)
      //
      for (pdbNamespace::memvec::const_iterator m = mv.begin();
	   m != mv.end();
	   ++m) {
	const pdbItem *memitem = static_cast<const pdbItem *>(*m);

	//
	// i think this has got to be the absolute worst thing in PDT
	// so far.  Nothing *NOTHING* gives any hint that we need to
	// cast across from pdbItem to pdbGroup and pdbClass.  In fact,
	// nothing allows us to look at a pdbItem and figure out what
	// it should be cast to.  This is bad.  very bad.
	//
	if (((const pdbGroup *)memitem)->kind() == pdbItem::GR_CLASS) {
	  outputClass((const pdbClass *)memitem,os,depth+1);
	}
      }
      
      //
      // close scope tag
      //
      indent(os,depth);
      os << "</scope>" << endl;
    }    
  }

  return 0;
}

/**
 * This method outputs all fortran modules
 */
int outputModules(PDB::modulevec& mv, ostream& os, int depth) {
  //
  // we're handed a module vector, so go through it and dump each module
  //
  for (PDB::modulevec::const_iterator m = mv.begin();
       m != mv.end();
       ++m) {
    const pdbModule *cur = static_cast<const pdbModule*>(*m);
    pdbModule::modfuncvec mfv = cur->funcMembers();

    //
    // make sure the module isn't empty.  we don't like
    // empty scopes.
    //
    if (mfv.begin() != mfv.end()) {

      //
      // comment block
      //
      os << endl;
      indent(os,depth);
      os << "<!--" << endl;
      indent(os,depth);
      os << "  - Module: " << cur->fullName() << endl;
      indent(os,depth);
      os << "  -->" << endl;
      
      //
      // begin scope, we know TYPE is fmodule
      //
      indent(os,depth);
      os << "<scope name=\"" << cur->fullName() << "\" kind=\"module\">"
         << endl;

      //
      // dump out user defined types
      //
      pdbGroup::fieldvec gfv = cur->dataMembers();
      for (pdbGroup::fieldvec::const_iterator f = gfv.begin();
           f != gfv.end();
           ++f) {
        const pdbGroupField *c = static_cast<const pdbGroupField*>(*f);

        // get the type of the group field
        const pdbType *ct = c->type();
        if (ct != NULL) {
          //
          // now, get the group for this type.  If it is NULL, then
          // this is not a user defined type.  If it is non-null, we
          // process it.
          //
          const pdbGroup *cg = ct->isGroup();
          if (cg != NULL)
            outputFUserType(cg,os,depth+1);
         }
      }

      // go through the module function members
      for (pdbModule::modfuncvec::const_iterator f = mfv.begin();
           f != mfv.end();
           ++f) {
        const pdbModFunc *cur = static_cast<const pdbModFunc*>(*f);
        
        //
        // dump the routine
        //
        outputFRoutine(cur->func(),os,depth+1);
      }

      //
      // close scope
      //
      indent(os,depth);
      os << "</scope>" << endl;
    }
  }
  
  return 0;
}

/**
 * This method outputs all fortran types
 */
int outputFTypes(PDB::typevec& tv, ostream& os, int depth) {
  bool dumped = false;
  bool printedScope = false;

  for (PDB::typevec::const_iterator t = tv.begin();
       t != tv.end();
       ++t) {
    const pdbType *cur = static_cast<const pdbType*>(*t);
    //
    // check if already dumped
    //
    for (unsigned int i = 0; i < dumpedTypes.size(); i++) {

      // we only deal with ones that isGroup() != NULL
      if (cur->isGroup() != NULL) {
        if ((string)dumpedTypes[i] ==
            cur->isGroup()->fullName()) {
          dumped = true;
          break;
        }
      } else {
        if ((string)dumpedTypes[i] == cur->fullName()) {
          dumped = true;
          break;
        }
      }
    }

    //
    // don't accidentally fall into printing scope if all we know about
    // are intrinsic types...
    //
    if (cur->isGroup() == NULL) dumped = true;

    // 
    // dumped == false means we really have at least one non-trivial
    // user defined type, so we dump it.
    //
    if (dumped == false) {
      if (printedScope == false) {
        indent(os,depth);
	//
	// NOTE: maybe here interface is implicit?
	//
        os << "<scope name=\"types_not_in_modules\" kind=\"global\">" <<
          endl;
        printedScope = true;
      }
      
      const pdbGroup *cg = cur->isGroup();

      //
      // this if should never fail thanks to the IF above.
      //
      if (cg != NULL)
        outputFUserType(cg,os,depth+1);

      // 
      // remember this type
      //
      dumpedTypes.push_back(cur->fullName());
    }

    // reset flag
    dumped = false;
  }

  //
  // close scope element if we opened it.
  //
  if (printedScope == true) {
    indent(os,depth);
    os << "</scope>" << endl;
  }

  return 0;
}

/**
 * This method outputs all fortran routines
 */
int outputFRoutines(PDB::froutinevec& frv, ostream& os, int depth) {
  bool dumped = false;
  bool printedScope = false;

  // iterate over each routine, outputting the XML for each
  for (PDB::froutinevec::const_iterator f = frv.begin();
       f != frv.end();
       ++f) {
    const pdbFRoutine *cur = static_cast<const pdbFRoutine*>(*f);

    //
    // check if it has already been dumped...
    //
    for (unsigned int i = 0; i < dumpedRoutines.size(); i++) {
      if ((string)dumpedRoutines[i] == cur->fullName()) {
        dumped = true;
        break;
      }
    }

    //
    // if a method has a null signature, we definitely don't dump it.
    // so pretend it has already been dumped.
    //
    const pdbType *sig = cur->signature();
    if (sig == NULL)
      dumped = true;

    if (dumped == false) {
      if (printedScope == false) {
        indent(os,depth);
        os << "<scope name=\"F90Global\" kind=\"global\">" <<
          endl;
        printedScope = true;
      }
      outputFRoutine(cur,os,depth+1);
      dumpedRoutines.push_back(cur->fullName());
    }

    dumped = false;
  }

  if (printedScope == true) {
    indent(os,depth);
    os << "</scope>" << endl;
  }
  
  return 0;
}

/**
 * dump classes from the given PDB::classvec that have not been dumped
 * yet.
 */
int outputClasses(PDB::classvec &cv, ostream &os, int depth) {
  bool dumped = false;
  bool printedScope = false;

  // iterate over each class, outputting the XML for each
  for (PDB::classvec::const_iterator c = cv.begin();
       c != cv.end();
       ++c) {
    const pdbClass *cur = static_cast<const pdbClass*>(*c);

    //
    // check if it has already been dumped...
    //
    for (unsigned int i = 0; i < dumpedClasses.size(); i++) {
      if ((string)dumpedClasses[i] == cur->fullName()) {
        dumped = true;
        break;
      }
    }

    if (dumped == false) {
      if (printedScope == false) {
        indent(os,depth);
        os << "<scope name=\"GlobalClasses\" kind=\"global\">" <<
          endl;
        printedScope = true;
      }
      outputClass(cur,os,depth+1);
      dumpedClasses.push_back(cur->fullName());
    }

    dumped = false;
  }

  if (printedScope == true) {
    indent(os,depth);
    os << "</scope>" << endl;
  }
  
  return 0;
}

/**
 * XML Generator Main : start things up.
 */
int main(int argc, char *argv[]) {
  bool    errflag = argc < 3;
  filebuf outfb;
  bool    forceC = false;

  //
  // really braindead argument processing going on here...
  if (errflag) {
    cerr << "usage: XML_Gen [-c] outfile.xml " <<
      "[file1.pdb file2.pdb ... filen.pdb]" << endl << endl;
    cerr << "       -c : optional flag.  forces generator to assume C, not C++." << endl << endl;
    return 0;
  }
  
  int startArg = 1;
  if (strcmp(argv[1],"-c") == 0) {
    startArg = 2;
    forceC = true;
    cerr << "DEBUG: XML_Gen forcing C." << endl;
  }

  // open output XML file
  outfb.open(argv[startArg],ios::out);
  ostream os(&outfb);
  ostream *osp = &os;

  startArg++;

  // doctype
  os << "<?xml version=\"1.0\"?>" << endl;
  os << "<!DOCTYPE library SYSTEM \"mapping.dtd\">" << endl;
  os << endl;

  //
  // iterate over all PDB files, each being considered a library of its
  // own.
  //
  for (int i = startArg; i < argc; i++) {    
    // create a new ductape PDB object
    PDB pdb(argv[i]);

    // reset the language flag
    libraryLanguage = (language_t)0;

    // a little header comment in the XML file telling source info
    // and creation info.
    os << "<!-- " << argv[startArg-1] << " : automatically generated XML -->" << endl;
    os << "<!-- Source PDB file : " << argv[i] << " -->" << endl;
    
    // output the outermost tag for the library
    os << "<library ";
    
    //
    // determine the language.  based on this we will call different
    // routines later to dump the XML.
    //
    if (forceC) {
      libraryLanguage = LANG_C;
      os << "lang=\"c\" ";
    } else {
      switch (pdb.language()) {
      case PDB::LA_C:
        libraryLanguage = LANG_C;
        os << "lang=\"c\" ";
        break;
      case PDB::LA_CXX:
      case PDB::LA_C_or_CXX:
        libraryLanguage = LANG_CXX;
        os << "lang=\"cxx\" ";
        break;
      case PDB::LA_FORTRAN:
        libraryLanguage = LANG_FORTRAN;
        os << "lang=\"fortran\" ";
        break;
      default:
        os << "lang=\"unknown\" ";
        cerr << "Error: unknown language in PDB file (" << pdb.language() <<
          ")" << endl;
      }
    }
    
    // 
    // if language wasn't set, something is wrong so we bail out.  the OS
    // file will contain a partially completed <LIBRARY tag, but it was
    // obviously messed up from the beginning anyways - so who cares?
    //
    if (libraryLanguage == 0) {
      cerr << "Error: PDB file did not contain a valid language identifier."
	   << endl;
      exit(0);
    }

    //
    // get version from pdb object -- I think this is PDT version, but ...
    //
    os << "version=\"" << pdb.version() << "\" ";
    
    //
    // name the library after the first filename
    //
    PDB::filevec fv = pdb.getFileVec();
    os << "name=\"" << (fv[0])->name() << "\">" << endl << endl;
    
    // ************************
    //  DEAL WITH FORTRAN 
    // ************************

    if (libraryLanguage == LANG_FORTRAN) {
      // dump fortran modules
      //
      os << "<!-- fortran module scopes -->" << endl;
      
      outputModules(pdb.getModuleVec(),*osp,1);
      
      // 
      // dump routines not in modules
      //
      os << "<!-- functions NOT contained in a scope -->" << endl;
      outputFRoutines(pdb.getFRoutineVec(),*osp,1);
      os << endl;
      
      os << "<!-- types NOT contained in a scope -->" << endl;
      outputFTypes(pdb.getTypeVec(),*osp,1);
      os << endl;
    } else if (libraryLanguage == LANG_C || libraryLanguage == LANG_CXX) {
      // ************************
      //  DEAL WITH C and C++
      // ************************

      if (libraryLanguage == LANG_CXX) {
	// first deal with the C++ side of this inbred language family...
	//
	os << endl << "<!-- c++ namespaces and classes -->" << endl;

	outputNamespaces(pdb.getNamespaceVec(),*osp,1);
      }
      
      // now classes NOT in namespaces
      // (this means structs too...)
      os << endl << "<!-- c++ classes not contained in namespaces and C structs -->" 
         << endl;
      
      outputClasses(pdb.getClassVec(),*osp,1);

      // what is left is the global C++ stuff outside classes and such, and
      // the C side of things.

      os << endl << "<!-- C and C++ routines not contained elsewhere --> "
         << endl;
      outputCRoutines(pdb.getCRoutineVec(),*osp,1);

    } else {
      cerr << "Big ol' fatal error: This should REALLY not happen.\n" << endl;
      exit(0);
    }

    // close the library tag
    os << endl << "</library>" << endl << endl;
  }

  // close output XML file
  outfb.close();

  return 0;
}
